/*
 * Copyright (c) 2021,2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jul 27 2021
 */
#include "acftsens.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <fcntl.h>
#include <unistd.h>

#include "ftsens_c_types.h"
#include "codels.h"


/* --- Activity connect ------------------------------------------------- */

/** Validation codel fts_check_rate of activity connect.
 *
 * Returns genom_ok.
 * Throws ftsens_e_nodev, ftsens_e_baddev, ftsens_e_inval.
 */
genom_event
fts_check_rate(uint32_t rate, const genom_context self)
{
  return rate < 1 ? ftsens_e_inval(self) : genom_ok;
}


/* --- Function set_frame_offset ---------------------------------------- */

/** Codel fts_set_offset of function set_frame_offset.
 *
 * Returns genom_ok.
 */
genom_event
fts_set_offset(double x, double y, double z, double yaw, double pitch,
               double roll, double frame[36],
               const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  fts_get_frame(x, y, z, yaw, pitch, roll, frame);
  return genom_ok;
}


/* --- Function log ----------------------------------------------------- */

/** Codel fts_log_start of function log.
 *
 * Returns genom_ok.
 * Throws ftsens_e_sys.
 */
genom_event
fts_log_start(const char path[64], uint32_t decimation,
              ftsens_log_s **log, const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return ftsens_e_sys_error(path, self);

  if (write(fd, ftsens_log_header "\n", sizeof(ftsens_log_header)) < 0)
    return ftsens_e_sys_error(path, self);

  if ((*log)->req.aio_fildes >= 0) {
    close((*log)->req.aio_fildes);

    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->req.aio_fildes = fd;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = decimation < 1 ? 1 : decimation;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel fts_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
fts_log_stop(ftsens_log_s **log, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (*log && (*log)->req.aio_fildes >= 0)
    close((*log)->req.aio_fildes);
  (*log)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel fts_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
fts_log_info(const ftsens_log_s *log, uint32_t *miss, uint32_t *total,
             const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  *miss = *total = 0;
  if (log) {
    *miss = log->missed;
    *total = log->total;
  }
  return genom_ok;
}
