/*
 * Copyright (c) 2021,2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jul 27 2021
 */
#include "acftsens.h"

#include <sys/time.h>
#include <err.h>
#include <pthread.h>
#include <stdatomic.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "ftsens_c_types.h"
#include "codels.h"

#include <icub/icub.h>


static void *		fts_reader(void *data);
static atomic_bool	fts_async_reader;


/* --- Task read -------------------------------------------------------- */


/** Codel fts_init of task read.
 *
 * Triggered by ftsens_start.
 * Yields to ftsens_pause_read.
 */
genom_event
fts_init(ftsens_ids *ids, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  /* init device */
  ids->hw = malloc(sizeof(*ids->hw));
  if (!ids->hw) return ftsens_e_sys_error("malloc", self);
  *ids->hw = (ftsens_hw){
    .dev = NULL,
    .node = NULL
  };

  /* default to no calibration */
  ids->calibration = (struct ftsens_ids_calibration_s){
    .G = {0},
    .b = {0}
  };
  fts_set_offset(0., 0., 0., 0., 0., 0., ids->frame, self);

  /* init data */
  ids->data = malloc(sizeof(*ids->data));
  if (!ids->data) return ftsens_e_sys_error("malloc", self);
  *ids->data = (strain2_data){};

  /* init logging */
  ids->log = malloc(sizeof(*ids->log));
  if (!ids->log) return ftsens_e_sys_error("malloc", self);
  *ids->log = (ftsens_log_s){
    .req = {
      .aio_fildes = -1,
      .aio_offset = 0,
      .aio_buf = ids->log->buffer,
      .aio_nbytes = 0,
      .aio_reqprio = 0,
      .aio_sigevent = { .sigev_notify = SIGEV_NONE },
      .aio_lio_opcode = LIO_NOP
    },
    .pending = false, .skipped = false,
    .decimation = 1, .missed = 0, .total = 0
  };

  return ftsens_pause_read;
}


/** Codel fts_read of task read.
 *
 * Triggered by ftsens_read.
 * Yields to ftsens_export, ftsens_pause_read.
 */
genom_event
fts_read(ftsens_hw **hw, strain2_data **data,
         const genom_context self)
{
  if (!(*hw)->dev) return ftsens_pause_read;

  if (icub_receive_and_decode((*hw)->dev, 500) < 0) {
    warn("error reading can bus");
    fts_disconnect_start(hw, self);
    return ftsens_pause_read;
  }

  if (icub_get_strain2_data((*hw)->node, *data)) {
    warn("error copying data");
    fts_disconnect_start(hw, self);
    return ftsens_pause_read;
  }

  return ftsens_export;
}


/** Codel fts_export of task read.
 *
 * Triggered by ftsens_export.
 * Yields to ftsens_read.
 */
genom_event
fts_export(const strain2_data *data, const double frame[36],
           ftsens_log_s **log, const ftsens_wrench *wrench,
           const genom_context self)
{
  struct timeval tv;
  double w[6];

  gettimeofday(&tv, NULL);

  /* apply transformation matrix */
  fts_get_wrench(data, frame, w);

  /* publish */
  *wrench->data(self) = (or_wrench_estimator_state){
    .ts = { .sec = tv.tv_sec, .nsec = tv.tv_usec * 1000 },
    .intrinsic = true,

    .force = {
      ._present = true,
      ._value = { .x = w[0], .y = w[1], .z = w[2] }
    },
    .force_cov = { ._present = false },

    .torque = {
      ._present = true,
      ._value = { .x = w[3], .y = w[4], .z = w[5] }
    },
    .torque_cov = { ._present = false }
  };
  wrench->write(self);

  /* log */
  if ((*log)->req.aio_fildes >= 0) {
    (*log)->total++;
    if ((*log)->total % (*log)->decimation == 0) {
      if ((*log)->pending) {
        if (aio_error(&(*log)->req) != EINPROGRESS) {
          (*log)->pending = false;
          if (aio_return(&(*log)->req) <= 0) {
            warn("log");
            close((*log)->req.aio_fildes);
            (*log)->req.aio_fildes = -1;
          }
        } else {
          (*log)->skipped = true;
          (*log)->missed++;
        }
      }
    }

    if ((*log)->req.aio_fildes >= 0 && !(*log)->pending) {
      (*log)->req.aio_nbytes = snprintf(
        (*log)->buffer, sizeof((*log)->buffer),
        "%s" ftsens_log_line "\n",
        (*log)->skipped ? "\n" : "",
        (uint64_t)tv.tv_sec, (uint32_t)tv.tv_usec * 1000,
        w[0], w[1], w[2], w[3], w[4], w[5]);

      if (aio_write(&(*log)->req)) {
        warn("log");
        close((*log)->req.aio_fildes);
        (*log)->req.aio_fildes = -1;
      } else
        (*log)->pending = true;

      (*log)->skipped = false;
    }
  }

  return ftsens_read;
}


/** Codel fts_stop of task read.
 *
 * Triggered by ftsens_stop.
 * Yields to ftsens_ether.
 */
genom_event
fts_stop(ftsens_log_s **log, ftsens_hw **hw, const genom_context self)
{
  fts_log_stop(log, self);
  if (*log) free(*log);

  fts_disconnect_start(hw, self);

  return ftsens_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel fts_connect_start of activity connect.
 *
 * Triggered by ftsens_start.
 * Yields to ftsens_ether.
 * Throws ftsens_e_nodev, ftsens_e_baddev, ftsens_e_inval.
 */
genom_event
fts_connect_start(const char iface[64], uint32_t rate, ftsens_hw **hw,
                  ftsens_ids_calibration_s *calibration,
                  const genom_context self)
{
  pthread_t reader;
  int i, j, s;
  uint8_t ms = 1000 / rate;

  /* drop any pending connection and start over */
  fts_disconnect_start(hw, self);

  (*hw)->dev = icub_init(iface);
  if (!(*hw)->dev) {
    warn("icub_init(%s)", iface);
    return ftsens_e_nodev(self);
  }

  /* probe device until one is found, with a timeout */
  for(i = 0; i < 10; i++) {
    icub_discover((*hw)->dev);
    for(j = 0; j < 10; j++) {
      s = icub_receive_and_decode((*hw)->dev, 100);
      if (s < 0) goto fail;
      if (s <= 0) break;
    }

    if (icub_find_node((*hw)->dev, ICUB_BOARD_STRAIN2) != NULL)
      break;
  }

  (*hw)->node = icub_find_node((*hw)->dev, ICUB_BOARD_STRAIN2);
  if ((*hw)->node == NULL) {
    warnx("can't find strain2 node");
    goto fail;
  }
  warnx("found strain2 as node %d\n", icub_can_address((*hw)->node));

  /* create async reader for handling replies during initialization */
  pthread_create(&reader, NULL, fts_reader, *hw);

  /* get information */
  icub_sensor_get_firmware_version((*hw)->node);
  icub_get_serial_number((*hw)->node);

  icub_sync((*hw)->node);
  icub_dump((*hw)->dev);

  /* get calibration */
  icub_regulationset_get((*hw)->node, ICUB_REGSET_VALIDITY_TEMPORARY);
  icub_set_calib_tare((*hw)->node,
                      ICUB_CALIB_MODE_EVERY_CHANNEL_NEGATIVE_OF_DAC,
                      ICUB_REGSET_INUSE, 0, 0);
  for (i = 0; i < 6; i++) {
    icub_sensor_get_fullscale((*hw)->node, ICUB_REGSET_INUSE, i);
    icub_get_calib_tare((*hw)->node, ICUB_REGSET_INUSE, i);
    icub_amplifier_gainoffset_get((*hw)->node, ICUB_REGSET_INUSE, i);

    for (j = 0; j < 6; j++)
      icub_get_matrix_rc((*hw)->node, ICUB_REGSET_INUSE, i, j);
  }

  icub_sync((*hw)->node);
  {
    strain2_data data;
    icub_get_strain2_data((*hw)->node, &data);
    for (i = 0; i < 6*6; i++)
      calibration->G[i] = data.matrix[i/6][i%6];
    for (i = 0; i < 6; i++)
      calibration->b[i] = data.tare[i];
  }

  /* set up data streaming */
  icub_sensor_set_can_data_rate((*hw)->node, ms);
  icub_sensor_set_txmode((*hw)->node, ICUB_STRAIN_MODE_TX_CALIBRATED);

  /* terminate async reader */
  fts_async_reader = false;
  pthread_join(reader, NULL);

  return ftsens_ether;

fail:
  icub_end((*hw)->dev);
  (*hw)->dev = NULL;
  return ftsens_e_baddev(self);
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel fts_disconnect_start of activity disconnect.
 *
 * Triggered by ftsens_start.
 * Yields to ftsens_ether.
 */
genom_event
fts_disconnect_start(ftsens_hw **hw, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (!(*hw)->dev) return ftsens_ether;

  /* stop data transmission */
  icub_sensor_set_txmode((*hw)->node, ICUB_STRAIN_MODE_TX_NONE);

  /* ack messages */
  while (icub_receive_and_decode((*hw)->dev, 500) > 0);

  /* close connection */
  icub_end((*hw)->dev);
  (*hw)->dev = NULL;
  (*hw)->node = NULL;
  warnx("disconnected from device");

  return ftsens_ether;
}


/* --- Activity set_zero ------------------------------------------------ */

/** Codel fts_zero_start of activity set_zero.
 *
 * Triggered by ftsens_start.
 * Yields to ftsens_loop.
 * Throws ftsens_e_nodev.
 */
genom_event
fts_zero_start(int32_t *n, double ib[6], const genom_context self)
{
  (void)self; /* -Wunused-parameter */
  int i;

  *n = 0;
  for(i=0; i<6; i++) ib[i] = 0.;

  return ftsens_loop;
}

/** Codel fts_zero_loop of activity set_zero.
 *
 * Triggered by ftsens_loop.
 * Yields to ftsens_pause_loop, ftsens_set.
 * Throws ftsens_e_nodev.
 */
genom_event
fts_zero_loop(const ftsens_hw *hw, const strain2_data *data,
              int32_t samples, double ib[6], int32_t *n,
              const ftsens_ids_calibration_s *calibration,
              const genom_context self)
{
  double adc[6];
  int i;

  /* this is called twice per sample, because of the read + export codels */

  if (!hw->dev) return ftsens_e_nodev(self);
  if ((*n)++ >= 2 * samples) return ftsens_set;

  fts_get_adc(data, calibration, adc);
  for(i = 0; i < 6; i++)
    ib[i] -= adc[i] / samples / 2;

  return ftsens_pause_loop;
}

/** Codel fts_zero_set of activity set_zero.
 *
 * Triggered by ftsens_set.
 * Yields to ftsens_ether.
 * Throws ftsens_e_nodev.
 */
genom_event
fts_zero_set(const ftsens_hw *hw, const double ib[6], double b[6],
             const genom_context self)
{
  (void)self; /* -Wunused-parameter */
  int i;

  for(i = 0; i < 6; i++) {
    icub_set_calib_tare(
      hw->node,
      ICUB_CALIB_MODE_SET_CHANNEL_WITH_VALUE, ICUB_REGSET_INUSE,
      i, ib[i]);
    b[i] = ib[i];
  }

  return ftsens_ether;
}




/* --- fts_reader ---------------------------------------------------------- */

static void *
fts_reader(void *data)
{
  ftsens_hw *hw = data;

  fts_async_reader = true;
  while (fts_async_reader) {
    if (icub_receive_and_decode(hw->dev, 500) < 0) {
      warn("error reading can bus");
      break;
    }
  }

  return NULL;
}
